#include <stdlib.h>

#include <iostream>
using namespace std;

#include "../syssoft_linux.cpp"

const bool LINKS = false;
const bool RECHTS = true;

class Esszimmer {
private:
	int n;

	struct Platz {
		int frei;
		Semaphore *staebchen;
	};

	Platz *tisch;
	Semaphore *haegar;
	Semaphore *mutex;
public:
	Esszimmer ( int p_n ) : n(p_n) {
		tisch = new Platz[n];
		for (int i=0; i<n; i++) {
			tisch[i].frei = 0;
			tisch[i].staebchen = new Semaphore(1);
		};
		haegar = new Semaphore(n);
		mutex = new Semaphore(1);
	};
	~Esszimmer () {
		for (int i=0; i<n; i++)
			delete tisch[i].staebchen;
		delete [] tisch;
	};

	int suche_freien_platz () {
		haegar->P();
		mutex->P();
		for (int i=0; i<n; i++)
			if (tisch[i].frei == 0) {
				tisch[i].frei++;
				mutex->V();
				return i;
			}
		cerr << "Hier sollten wir nicht sein." << endl;
	};

	void verlasse_platz ( int p ) {
		mutex->P();
		tisch[p].frei--;
		mutex->V();
		haegar->V();
	};

	void nimm_staebchen ( int p, bool wo ) {
		int i = (wo == LINKS) ? p : (p+1) % n;
		tisch[i].staebchen->P();
	};

	void lege_staebchen_weg ( int p, bool wo ) {
		int i = (wo == LINKS) ? p : (p+1) % n;
		tisch[i].staebchen->V();
	};

};

struct Weihnachtsmann_DNA {
	Esszimmer *zimmer;
	int ident;
};

void ausgabe ( int kennung, const char *comment, int i ) {
	static Semaphore mutex(1);

	mutex.P();
	cerr << "WM[" << kennung << "] " << comment << " " << i << endl;
	mutex.V();
}

void *Weihnachtsmann ( void *arg ) {
	Weihnachtsmann_DNA *dna = reinterpret_cast<Weihnachtsmann_DNA *>(arg);
	Esszimmer *ez = dna->zimmer;
	int name = dna->ident;
	delete dna;

	while (true) {
		// Weihnachten ueben
//		usleep(random() % 100000);
		// Hunger !!
		ausgabe(name,"Suche freien Platz",0);
		int p = ez->suche_freien_platz();
		ausgabe(name,"Bekomme Platz",p);
		ausgabe(name,"Nehme linkes Staebchen",0);
		ez->nimm_staebchen(p,LINKS);
		ausgabe(name,"Nehme rechtes Staebchen",0);
		ez->nimm_staebchen(p,RECHTS);
		ausgabe(name,"ESSEN",0);
//		usleep(random() % 100000);
		ausgabe(name,"Lege rechtes Staebchen weg",0);
		ez->lege_staebchen_weg(p,RECHTS);
		ausgabe(name,"Lege linkes Staebchen weg",0);
		ez->lege_staebchen_weg(p,LINKS);
		ausgabe(name,"Verlasse PLatz",p);
		ez->verlasse_platz(p);
	}
}

int main ( int ac, char **av ) {
	if (ac != 3) {
		cerr << "usage: " << av[0] << " <#Maenner> <#Plaetze>" << endl;
		exit(0);
	}

	int anzahl_wmaenner = atoi(av[1]);
	int anzahl_plaetze = atoi(av[2]);

	Esszimmer ez(anzahl_plaetze);

	Thread **wms = new (Thread *)[anzahl_wmaenner];
	for (int wm = 0; wm < anzahl_wmaenner; wm++) {
		wms[wm] = new Thread(Weihnachtsmann);
		Weihnachtsmann_DNA *dna = new Weihnachtsmann_DNA;
		dna->zimmer = &ez;
		dna->ident = wm;
		wms[wm]->run(reinterpret_cast<void *>(dna));
	}

	wms[0]->join();

	return 24;
}

